#include "LuaUtil.h"
#include <iostream>
#include <string>

using namespace std;

template <typename T>
void printArray(const vector<T>& array)
{
  for(size_t i = 0; array.size() > i; ++i)
  {
    cout << array[i] << " ";
  }
}

struct S
{
  void print() const
  {
    cout << "User Type: ";
    cout << "text[" << text << "] ";
    cout << "values[";
    printArray(values);
    cout << "]" << endl;
  }
  string text;
  vector<int> values;
  bool operator < (const S& other) const
  {
    return text < other.text;
  }
};

template<> void toLua(lua_State* L, const S& arg)
{
  // stack:
  lua_newtable(L); // stack: table
  toLua(L, arg.text); // stack: table text
  lua_setfield(L, -2, "text"); // stack: table
  toLua(L, arg.values); // stack: table values
  lua_setfield(L, -2, "values"); // stack: table
}

template<> bool fromLua(lua_State* L, int index, S& arg)
{
  // stack:
  if(lua_istable(L, index))
  {
    string text;
    vector<int> values;
    if(fromLua(L, index, "text", text) &&
       fromLua(L, index, "values", values))
    {
      arg.text = text;
      arg.values = values;

      return true;
    }
  }

  return false;
}

int func(lua_State* L)
{
  //   
  toLua(L, true); // args 1
  toLua(L, 10); // args 2
  toLua(L, 2.3); // args 3
  toLua(L, string("xyz")); // args 4

  // 
  toLua(L, vector<int>(3, 100)); // args 5

  // 
  map<string, int> m;
  m["one"]= 1;
  m["two"]= 2;
  m["three"]= 3;
  toLua(L, m); // args 6

  //  
  S s;
  s.text = "userType";
  s.values.push_back(200);
  s.values.push_back(300);
  s.values.push_back(400);
  toLua(L, s); // args 7

  //   
  vector<S> vs;
  s.text = "text1";
  s.values = vector<int>(3, 1);
  vs.push_back(s);
  s.text = "text2";
  s.values = vector<int>(3, 2);
  vs.push_back(s);
  s.text = "text3";
  s.values = vector<int>(3, 3);
  vs.push_back(s);
  toLua(L, vs); // args 8

  //   
  //   
  //     
  map<S, string> ms;
  //  ,   ,
  //        <
  s.text = "text4";
  s.values = vector<int>(3, 4);
  ms[s] = "value4";
  s.text = "text5";
  s.values = vector<int>(3, 5);
  ms[s] = "value5";
  s.text = "text6";
  s.values = vector<int>(3, 6);
  ms[s] = "value6";
  toLua(L, ms); // args 9

  return 9;
}

void testFromLua(lua_State* L)
{
  //   
	bool boolVal;
	int intVal;
	double doubleVal;
  string text;
  fromLua(L, 0, "bool", boolVal);
  fromLua(L, 0, "int", intVal);
  fromLua(L, 0, "double", doubleVal);
  fromLua(L, 0, "text", text);
  cout << "Boolean:" << boolVal << endl;
  cout << "Int:" << intVal << endl;
  cout << "Double:" << doubleVal << endl;
  cout << "Text:" << text << endl;

  // 
  vector<int> array;
  fromLua(L, 0, "array", array);
  cout << "Array:" << endl;
  printArray(array);
  cout << endl;

  // 
  map<string, int> m;
  fromLua(L, 0, "map", m);
  cout << "Map:" << endl;
  for(map<string, int>::const_iterator i = m.begin(); m.end() != i; ++i)
  {
    cout << "[" << i->first << "] = " << i->second << endl;
  }

  //  
  S s;
  fromLua(L, 0, "userType", s);
  s.print();

  //   
  vector<S> vs;
  fromLua(L, 0, "userTypeArray", vs);
  cout << "User Type Array:" << endl;
  for(size_t i = 0; vs.size() > i; ++i)
  {
    vs[i].print();
  }

  //   
  map<string, S> ms;
  fromLua(L, 0, "userTypeMap", ms);
  cout << "User Type Map:" << endl;
  for(map<string, S>::const_iterator i = ms.begin(); ms.end() != i; ++i)
  {
    cout << "[" << i->first << "] = ";
    i->second.print();
  }
}

int main()
{
  lua_State* L = lua_open();
  luaL_openlibs(L);
  lua_pushcfunction(L, func); // stack: function
  lua_setglobal(L, "func"); // stack:
  //   
  // if(luaL_dofile(L, "fromLuaTest.lua"))
  // if(luaL_dofile(L, "toLuaTest.lua"))
  {
    printf(lua_tostring(L, -1));
    exit(0);
  }
  //testFromLua(L);

	return 0;
}
